/*    Copyright (C) 2005 XIAO, Gang of Universite de Nice - Sophia Antipolis
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

	/* Userspace interface for Linux sysmask */

int errno;

#include "smkd.h"

struct smkd_request request;
struct smkd_answer ans;
#define givemask ans.smk
struct timespec sleept, remt;
int debug, new_confcache, new_passwd;
int thiscall;
int namet, namelen, pathlen, depth, backupcnt;
int thisusec;	/* microsecond; for internal benchmarking */
int benchmark=0;	/* this should be manually set to use benchmark */
time_t now, tokencheck;

	/* Values to be defined in smkdrc */
uid_t denyid=0;
int follow_dotdot=0;

char reqname[BACKUP_MAX][SMK_PATH_MAX+1];
	/* The hash checks what lower case letters should appear
	 * in the name. Very simple but highly efficient. */
int reqnamelen[BACKUP_MAX], reqhash[BACKUP_MAX];
int reqnamecnt, thisf;

#include "masktab.c"
#include "smklib.c"
#include "tokens.c"
#include "smkdlib.c"
#include "smkdconf.c"
#include "smkdproc.c"
#include "socket.c"

void do_debug(int req)
{
    int call;
    char *reqname, *callname;
    
    switch(req) {
	case SMKD_REQ_OPEN: {
	    switch(request.flags&3) {
		case 1:		reqname="open-r "; break;
		case 2:		reqname="open-w "; break;
		case 3:		reqname="open-rw "; break;
		default:	reqname="open-0 "; break;
	    }
	    break;
	}
	case SMKD_REQ_EXEC:	reqname="exec "; break;
	case SMKD_REQ_SOCKET:	reqname="socket "; break;
	case SMKD_REQ_CALL:	reqname="call "; break;
	case SMKD_REQ_TOKEN:	reqname="token "; break;
	/* case SMKD_REQ_TRIGGER:	reqname="trigger "; break; */
	default: return;
    }
    logprefix(""); 
    call=request.syscall&0xFFFF;
    if(req==SMKD_REQ_CALL || call==__NR_open) print(reqname);
    else {
	callname=NULL;
	if(call>=0 && call<NR_sysmasks) callname=syscall_names[call].name;
	if(callname==NULL || *callname==0) printu(call);
	else {print(callname); print(" ");}
    }
    if(benchmark) printu(getusec()-thisusec);
    print(request.fname); 
    print(" ans="); printhex(ans.ind>>16); printu(ans.ind&0xFFFF);
    print("\n");
    if(new_confcache) {
	print("	Confcache: ");
	printu(cccnt); print("tokens, ");
	printu((cache_end-confcache+1023)/1024); print("K.\n");
    }
}

void token_mask(int token,struct sysmask_struct *sm)
{
    int k,t;
    for(k=0;k<2;k++) {
	t=(token>>(k*16))&0xFFFF;
	if(t) getmask(t+(k<<16),sm);
    }
}

void check_fname(int type)
{
    long i, k, t, acc, ac2, aclock, tt;
    int islast;
    long usertoken;
    struct sysmask_struct sm;
    char *p1;

    ans.ind=0;
    acc=4; ac2=aclock=0; thisf=0; islast=1;
    if(type==conf_sock) goto start;
    if(type==conf_exec) goto start;
    if(type==conf_call) goto start;
    acc=request.flags&3;
    if(acc==0) goto out;
    if(!(request.sm2&SM2_ROPEN)) acc&=2;
    if(!(request.sm2&SM2_WOPEN)) acc&=1;
    if(!acc) {	/* trivially accepted */
	ans.ind=0; return;
    }
start:
    memset(&sm,0,sizeof(sm));
restart:
    ans.ind=(ans.ind&0xFFFF0000) | EACCES;
    if(request.token==0) goto out;
    if(thisf==reqnamecnt-1) {
	islast=1; thiscall=request.syscall&0xFFFF;
    }
    else {
	islast=0; thiscall=CALL_BACKUP;
    }
    tt=0; usertoken=0;
    for(k=0;k<2;k++) {
	t=(request.token>>(k*16))&0xFFFF; if(t==0) continue;
	ac2=acc&3; if(!islast && ac2) ac2=1;
	if(!getdef(t+(k<<16),type)) {	/* no matching: refuse. */
	    aclock |= ac2; goto out;
	}
	else {
		/* masks for intermediate chains are not added */
	    if(islast) ans.ind|=def2mask(&sm);
	    if(islast) switch(type) {
		case conf_exec: {
		    if(k==0 && defs_newuid) {
			i=str2uid(defs_newuid); if(i>=0) {
			    ans.uid=i; ans.ind|=SMKD_ANS_NEWUID;
			}
		    }
		    if(k==0 && defs_newgid) {
			i=str2gid(defs_newgid); if(i>=0) {
			    ans.gid=i; ans.ind|=SMKD_ANS_NEWGID;
			}
		    }
		    goto checkuid;
		}
		case conf_open: {
		    checkuid:
		    if(defs_checkuid && !(ans.ind&SMKD_ANS_CHECKUID)) {
			i=str2uid(defs_checkuid); if(i>0 && i!=request.uid) {
			    ans.uid=i; ans.ind|=SMKD_ANS_CHECKUID;
			}
		    }
		    break;
		}
		case conf_sock: {
		    break;
		}
	    }
	    
	    if(defs_access!=NULL) {
		if(type==conf_open) {
		    for(p1=defs_access; *p1; p1++) {
			if(*p1=='r') ac2&=2;
			if(*p1=='w') ac2&=1;
		    }
		}
		else {
		    if(wordchr(defs_access,"deny")) goto out;
		}
	    }
	    
	    if(islast && k==0 && defs_newusertoken)
	      usertoken=(str2int(defs_newusertoken)<<16)&0xFFFF0000;
	    if(islast && defs_newtoken) {	/* this must be the last one
				 * for it resets the defs structure. */
		i=(str2int(defs_newtoken)&0xFFFF)<<(k*16);
		if(i) {
		    addtoken(&tt,i); token_mask(i,&givemask);
		    ans.ind|=SMKD_ANS_NEWMASK;
		}
		else if(strcmp(defs_newtoken,"none")==0) {
		    ans.ind|=SMKD_ANS_NEWMASK;
		}
	    }
	}
	if(ac2) {aclock |= ac2; goto out;}
		/* retain old token and mask if no change */
	if(islast && !(ans.ind&SMKD_ANS_NEWMASK)) { 
	    tt|=t<<(k*16); if(k==0) token_mask(t,&givemask);
	}
    }
    if(usertoken && !(tt&0xFFFF0000)) {
	addtoken(&tt,usertoken); token_mask(usertoken,&givemask);
	ans.ind|=SMKD_ANS_NEWMASK;
    }
    ans.ind&=0xFFFF0000;
out:
    if(!islast && (!(ans.ind&0xFFFF) || (acc&3&~aclock))) {
	thisf++; goto restart;
    }
    if((ans.ind&(SMKD_ANS_ADDMASK|SMKD_ANS_NEWMASK)) == SMKD_ANS_ADDMASK) {
	givemask=sm;
    }
    if(givemask.token) addkill(givemask.token);
    givemask.token=tt;
    ans.opt=acc&3&~aclock;
}

void check_socket(void)
{
    if(sock2str()==0) {
	logreq("bad_socket ");
	return;
    }
    check_fname(conf_sock);
}

void check_call(int call)
{
    char nbuf[32];
    const char *comp[16];
    
    call&=0xFFFF;
    int2str(call,nbuf,sizeof(nbuf));
    comp[0]="/"; comp[1]=nbuf; comp[2]="/";
    comp[3]=comp[4]=NULL;
    if(call>=0 && call<NR_sysmasks) 
      comp[3]=syscall_names[call].name;
    if(makestr(request.fname,comp,sizeof(request.fname))) {
	ans.ind=ENAMETOOLONG; return;
    }
    process_pathname(1); check_fname(conf_call);    
}

int main(int argc,char *argv[])
{
    struct stat st;
    time_t newtime;
    pid_t pid;
    int i, k, req;

    logptr=logbuf; logbuf[0]=0;
    i=sysmask(SMK_VERSION,NULL,0);
    if(i!=SYSMASK_VERSION) {
	print(argv[0]); print(": wrong kernel sysmask version ");
	printu(i); print("\n"); tostderr();
	return 1;
    }
    pid=fork();
    if(pid) {
	sleept.tv_sec=0; sleept.tv_nsec=30000000;
	nanosleep(&sleept, &remt);
	return 0;
    }
    now=time(0);
    passwdtime=0; getpasswd(); rc();
    if(sysmask(SMK_SMKDREG,NULL,denyid)) {
	print("smkd: registration refused by kernel.\n");
	tostderr(); return -1;
    }
    confcache_init();
    get_syscall_names();
    sleept.tv_sec=0; sleept.tv_nsec=10000000;
    tokentime=tokencheck=0; k=killcnt=0;
    nice(-5);
    do {
	if(tokencheck<=now-STAT_DELAY) {
	    tokencheck=now; gettokens(); getpasswd();
	    if(stat(confdir_s,&st)==0) {
		if(st.st_mode&(S_IRWXG|S_IRWXO)) chmod(confdir_s,S_IRWXU);
		if(denyid && st.st_uid!=denyid) chown(confdir_s,denyid,denyid);
	    }
	}
	while(sysmask(SMK_SMKDREQ,&request,sizeof(request))==0) {
	    debug=new_confcache=new_passwd=0;
	    if(benchmark) thisusec=getusec(); 
	    memset(&ans,0,sizeof(ans)); reqnamecnt=0;
	    req=request.reqtype&0xFF;
		/* length overflow check */
	    if(request.reqlen<0 || request.reqlen>=SMK_PATH_MAX) {
		logreq("bad_reqlen ");
		ans.ind=EINVAL;
		goto reply;
	    }
	    request.fname[request.reqlen]=0;
	    if(req==SMKD_REQ_OPEN || req==SMKD_REQ_EXEC) {
		pathlen=request.reqlen; request.fname[pathlen]=0;
			/* path must be absolute */
		if(request.fname[0]!='/' ||
		   pathlen>=SMK_PATH_MAX || pathlen<=0) {
		    logreq("badfname ");
		    ans.ind=ENAMETOOLONG;
		    goto reply;
		}
		ans.ind=process_pathname(0);
		if(ans.ind) goto reply;
	    }
	    switch(req) {
		case SMKD_REQ_OPEN: {
		    check_fname(conf_open); break;
		}
		case SMKD_REQ_EXEC: {
		    check_fname(conf_exec); break;
		}
		case SMKD_REQ_SOCKET: {
		    check_socket(); break;
		}
		case SMKD_REQ_TOKEN: {
		    token_mask(request.token,&givemask);
		    if(givemask.token) addkill(givemask.token);
		    break;
		}
		case SMKD_REQ_LOG: {
		    call_log(request.syscall);
		    break;
		}
		case SMKD_REQ_CALL: {
		    check_call(request.syscall);
		    break;
		}
		case SMKD_REQ_TRIGGER: { /* not implemented: design questions. */
		    ans.ind=EINVAL;
		    break;
		}
		default: {
		    ans.ind=EINVAL;
		    break;
		}
	    }
	    reply: 
	    if(backupcnt>0 && follow_dotdot) ans.ind |= SMKD_ANS_HASDOTDOT;
	    if(ans.ind & SMKD_ANS_KILL) {
		ans.smk.active.m1=request.pid;
		ans.smk.active.m2=request.start_time;
	    }
	    sysmask(SMK_SMKDREP,&ans,sizeof(ans));
	    if((request.sm2&SM2_LOG) && (debug || (ans.ind&0xFFFF))) do_debug(req);
	    if(new_passwd) confcache_init();
	}
	k++; if((k&31)==0) {
	    newtime=time(0); if(newtime!=now) {
		now=newtime; checkkill(); writelog();
		if(cache_time > now) cache_time=now;
	    }
	}
	nanosleep(&sleept,&remt);
    } while(1);
    return 0;
}

