/*    Copyright (C) 2005 XIAO, Gang of Universite de Nice - Sophia Antipolis
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

	/* smkd: configuration file parser. */

#define defcnt (sizeof(defs)/sizeof(defs[0]))
#define cdefcnt (sizeof(cdef)/sizeof(cdef[0]))

struct {
    int cquota, tquota;
} usercquota[ACCOUNTLIM];

struct ccline {
    short int type, len;
    int hash;
    char *name, *def;
};

struct ccblock {
    struct ccline *lines;
};

struct cfp {
    time_t laststat;
    struct ccblock *blocks;
};

struct ccp {
    struct cfp mask, open, exec, socket, call, trigger;
};

struct {
    int token, user;
    struct ccp *cp;
    int pad;
} cctab[CONFCACHELIM];
int cccnt;

char confcache[CONFCACHESIZE];
char *cache_end;
time_t cache_time;
char confbuf[CONFLEN];

void confcache_init(void)
{
    cache_end=confcache; cccnt=0;
    memset(cctab,0,sizeof(cctab));
    memset(usercquota,0,sizeof(usercquota));
    cache_time=time(0);
}

	/* searches confcache for token and user. */
int search_ccache(int token, int user)
{
    int i1,i2,j,k;
    
    if(cccnt<=0) return -1;
    k=cctab[0].token-token; if(k==0) k=cctab[0].user-user;
    if(k==0) return k; if(k>0) return -1;
    k=cctab[cccnt-1].token-token; if(k==0) k=cctab[cccnt-1].user-user;
    if(k==0) return cccnt-1; if(k<0) return ~cccnt;
    for(i1=0,i2=cccnt-1;i2>i1+1;) {
	j=(i2+i1)/2;
	k=cctab[j].token-token; if(k==0) k=cctab[j].user-user;
	if(k==0) return j;
	if(k>0) {i2=j; continue;}
	if(k<0) {i1=j; continue;}	
    }
    return ~i2;
}

int cctab_insert(struct ccp *cp, int token, int user)
{
    int k;
    k=search_ccache(token,user);
    if(k>=0) {cctab[k].cp=cp; return k;}
    if(cccnt>=CONFCACHELIM) return -1;
    k=~k;
    memmove(cctab+k+1,cctab+k,(cccnt-k)*sizeof(cctab[0]));
    cctab[k].token=token; cctab[k].user=user;
    cctab[k].cp=cp;
    cccnt++; return k;
}

	/* align must be a power of 2 */
void *getccache(int size, int align)
{
    char *old_end;
    int test;
    if(size<=0) return cache_end;
    if(size+(cache_end-confcache)+align >= CONFCACHESIZE) return NULL;
    if(align>1) {
	test=(long) cache_end%align; if(test) cache_end+=align-test;
    }
    old_end=cache_end; cache_end+=size;
    return old_end;
}

	/* options field */
void options(void)
{
    int i, j, *ip;
    char *p;
    p=defs_option;
    if(!p || *p==0) return;
    if(*p=='@') {
	p++; ip=(int *) p; i=*ip;
    }
    else {	/* compilation */
	char *p1, *p2;
	i=0;
	for(p1=find_word_start(p); *p1; p1=find_word_start(p2)) {
	    p2=find_word_end(p1); if(*p2) *p2++=0;
	    j=search_list(fnametab,fnamecnt,sizeof(fnametab[0]),p1);
	    if(j>=0) i|=fnametab[j].val;
	}
	*p='@'; p++; ip=(int *) p; *ip=i;
    }
    ans.ind |= i&0xFF000000;
    if(i & NAME_DEBUG2) i |= NAME_DEBUG;
    if(i & NAME_DEBUG && (request.sm2&SM2_LOG)) {
	debug |= 1;
	if(i&NAME_DEBUG2) debug |= 2;
    }
    if(i & NAME_KEEPUID) {
	ans.uid=request.uid; ans.ind |= SMKD_ANS_NEWUID;
    }
}

	/* returns match_end if match, recursive */
int _testname(char *tester, char *name)
{
    char *p1, *p2;
    
    for(p1=tester, p2=name; *p1 && *p2; p1++, p2++) {
	if(*p1==*p2) continue;
	switch(*p1) {
	    case '?': {
		if(*p2=='/' || (*p2=='.' && (p2==name || p2[-1]=='/'))) return 0;
		else continue;
	    }
	    case '*': {	/* we do not allow '?' following '*' */
		p1++;
		if(*p1=='*') {
		    p1++; while(*p1=='*') p1++;
		    if(*p1==0) return 1;
		    rep1:
		    for(; *p2 && *p2!=*p1; p2++);
		    if(!*p2) return 0;
		    if(_testname(p1,p2)) return 1;
		    p2++; goto rep1;
		}
		else {
		    if(*p2=='.' && (p2==name || p2[-1]=='/')) return 0;
		    rep2:
		    for(; *p2 && *p2!=*p1 && *p2!='/'; p2++);
		    if(_testname(p1,p2)) return 1;
		    if(*p2=='/' || *p2==0) return 0;
		    p2++; goto rep2;
		}
	    }
	}
	return 0;
    }
    if(*p1==0) {
	if(*p2==0) return 1;
	if(p1>tester && p1[-1]=='/') return 1;
	else return 0;
    }
    else {	/* too short */
	if(*p1=='*' && (p1[1]==0 || (p1[1]=='*' && p1[2]==0))) return 1;
	else return 0;
    }
}

	/* returns 1 if match */
int testname(struct ccline *ln)
{
    long l;
    char *p, *req;
    char buf[SMK_PATH_MAX+17];
    
    p=ln->name; l=ln->len;
    req=reqname[thisf];
    if(l>=SMK_PATH_MAX || l<=0) return 0;
    if(*p=='/') {
	if(l==1) {	/* this matches everything */
	    ans.ind |= SMKD_ANS_FOLLOW;
	    return 1;
	}
	if(l==2 && p[1]=='.') { /* this matches "/" */
	    if(reqnamelen[thisf]==1 && req[0]=='/') return 1;
	    else return 0;
	}
    }
    if(ln->type==conftype_path2) {
	memmove(buf,p,l); buf[l]=0;
	if(!substit(buf)) return 0;
	p=find_word_start(buf); l=strlen(p);
    }
    if(l<=0) return 0;
    if(*p=='*' && p[1]!='*') {
	buf[0]='*'; memmove(buf+1,p,l+1); l++; p=buf;
    }
    return _testname(p, req);
}

	/* returns 0 or 1. */
int testuser(char *p)
{
    unsigned long list[LISTLIM];
    int i, n;

    if(*p!='@') {
	n=strlen(p); if(n<LINELEN_MIN-2) n=LINELEN_MIN-2;
	i=rlist(list,p,users,usercnt,sizeof(users[0]));
	if(i<LISTLIM && i*sizeof(list[0])+2<=n) { /* compilation */
	    p[0]='@'; p[1]=i;
	    memmove(p+2,list,i*sizeof(list[0]));
	}
	return rcheck(list, request.uid, i);
    }
    else {
	i=p[1]; i&=255; p+=2; 
	return rcheck((unsigned long *) p, request.uid, i);
    }
}

	/* returns 0 or 1. */
int testcall(char *p)
{
    unsigned long list[LISTLIM];
    int i, n;

    if(*p!='@') {
	n=strlen(p); if(n<LINELEN_MIN-2) n=LINELEN_MIN-2;
	i=rlist(list,p,triggers,trcnt,sizeof(triggers[0]));
	if(i<LISTLIM && i*sizeof(list[0])+2<=n) { /* compilation */
	    p[0]='@'; p[1]=i;
	    memmove(p+2,list,i*sizeof(list[0]));
	}
	return rcheck(list, thiscall, i);
    }
    else {
	i=p[1]; i&=255; p+=2; 
	return rcheck((unsigned long *) p, thiscall, i);
    }
}

	/* returns 0 or 1. */
int testrefuse(char *str)
{
    int i, k;
    char *p, *pp;
    int *ip;
    if(*str!='@') {	/* compile */
	i=0;
	for(p=find_word_start(str); *p; p=find_word_start(pp)) {
	    pp=find_word_end(p); if(*pp) *pp++=0;
	    k=search_list(fnametab,fnamecnt,sizeof(fnametab[0]),p);
	    if(k>=0) i|=fnametab[k].val;
	}
	str[0]='@'; p=str+1; ip=(int *) p; *ip=i;
    }
    else {
	p=str+1; ip=(int *) p; i=*ip;
    }
    if(namet & i) return 0;
    if(i & NAME_HARDLINK) ans.ind |= SMKD_ANS_NOHLINK;
    if(i & NAME_SYMLINK) ans.ind |= SMKD_ANS_NOSLINK;
    return 1;
}

	/* returns 0 or 1. */
int testmask(char *str, long tester)
{
    int *ip;
    char *p;
    struct sysmask_struct sm;
    
    p=str+1; ip=(int *) p;
    if(*str!='@') {
	memset(&sm,0,sizeof(sm));
	str2mask(str,&sm.active,sm.except);
	if(*str!='@') {	/* if str2mask does not store the compilation */
	    *str='@'; p=str+1; ip=(int *) p;
	    ip[0]=sm.active.m1; ip[1]=sm.active.m2;
	}
    }
    if((request.sm1^tester)&ip[0]) return 0;
    if((request.sm2^tester)&ip[1]) return 0;
    return 1;
}

int test_cond(struct ccline *l, int type)
{
    int i,negate;
    i=l->name[0]; if(i&0x80) negate=1; else negate=0;
    i&=127;
    if(i>=cdefcnt || !(cdef[i].type&type)) return 1;	/* nothing to do */
    switch(cdef[i].val) {
	case defname_backlim: {
	    i=str2int(l->def);
	    if(i>0 && i<backupcnt) return negate;
	    else return 1^negate;
	}
	case defname_call: {
	    return testcall(l->def)^negate;
	}
	case defname_depthlim: {
	    i=str2int(l->def);
	    if(i>0 && i<depth) return negate;
	    else return 1^negate;
	}
	case defname_maskset: {
	    return testmask(l->def, (long) -1)^negate;
	}
	case defname_maskunset: {
	    return testmask(l->def, 0)^negate;
	}
	case defname_namelim: {
	    i=str2int(l->def);
	    if(i>0 && i<namelen) return negate;
	    else return 1^negate;
	}
	case defname_pathlim: {
	    i=str2int(l->def);
	    if(i>0 && i<pathlen) return negate;
	    else return 1^negate;
	}
	case defname_refuse: {
	    return testrefuse(l->def)^negate;
	}
	case defname_uid: {
	    return testuser(l->def)^negate;
	}
    }
    return 1;
}

	/* check postconditions:
	 * returns 1 if OK. */
int readdef(struct ccline *l)
{
    int i, j;

    memset(&defs,0,sizeof(defs));
    for(i=0; l[i].type!=conftype_empty; i++) {
	if(l[i].type!=conftype_def) continue;
	j=l[i].name[0]&127; if(j>=cdefcnt) continue;
	if(!(cdef[j].type & DEFTYPE_POST)) {
	    defs[cdef[j].val]=l[i].def; continue;
	}
	if(test_cond(l+i,DEFTYPE_POST)==0) return 0;
    }
    options();
    return 1;
}

	/* Find definition corresponding to request file in confbuf
	 * Returns the definition portion, or NULL if not found. */
struct ccline *findf(struct cfp *cf)
{
    struct ccline *ln;
    int i,j,h,t,pass;

    if(!cf || !cf->blocks || !cf->blocks->lines) return NULL;	/* empty */
	/* content blocks */
    h=~reqhash[thisf];
    for(i=0;cf->blocks[i].lines;i++) {
	ln=cf->blocks[i].lines;
	if(ln->type==conftype_def) {	/* condition block */
	    if(readdef(ln)==0) return NULL;
	    else continue;
	}
	for(j=0;ln[j].name[0];j++) {
	    switch(ln[j].type) {
		case conftype_pre: {
		    if(!test_cond(ln+j,DEFTYPE_PRE)) goto nextblock;
		    break;
		}
		case conftype_path:
		case conftype_path2: {
		    t = ln[j].hash & h;
		    if(t) pass=0; else pass=testname(ln+j);
		    if(debug&2) {
			logprefix("debug ");
			if(pass) print("MATCH: ");
			else {
			    if(t) {
				print("no hash: "); printhex(t);
			    }
			    else print("no match: ");
			}
			print(ln[j].name); print(" ~ ");
			print(reqname[thisf]);
			print("\n");
		    }
		    if(pass) return ln+j+1;
		    else break;
		}
	    }
	}
	nextblock:;
    }
    return NULL;
}

	/* return 0 if error */
int cutconf(char buf[], struct cfp *cf)
{
    struct ccblock blk[CONFBLOCKS];
    struct ccline *cl;
    char *ln[CONFLINES];
    int type[CONFLINES];
    int i, l, inblock;
    int blocks, lines, negate;
    char *p1, *p2, *bp, *empty;
    
    
    blocks=lines=0; bp=cache_end;
    for(p1=find_word_start(buf); 
	*p1 && lines<CONFLINES-1 && bp<confcache+CONFCACHESIZE;
	p1=p2) {
	p2=find_line_end(p1); if(*p2) *p2++=0;
	p1=find_word_start(p1);
	switch(*p1) {
	    case '#': continue; /* comments */
	    case 0: {
		if(lines>0 && type[lines-1]==conftype_empty) continue;
		type[lines]=conftype_empty;
		ln[lines++]=bp;	*bp++=0;
		p2=find_word_start(p2);
		continue;
	    }
	    case '>': {
		type[lines]=conftype_pre; p1++;
		goto def;
	    }
	    case '*':
	    case '%':
	    case '/': {
		*find_word_end(p1)=0;
		if(strchr(p1,'%')==NULL) type[lines]=conftype_path;
		else type[lines]=conftype_path2;
		break;
	    }
	    default: {
		type[lines]=conftype_def;
		def:
		if(!myisalnum(*p1) && *p1!='_') continue; /* unknown name */
		if(strchr(p1,'=')==NULL) continue;
		break;
	    }
	}
	l=strlen(p1);
		/* cache overflow: need reinit. */
	if(l + (bp-confcache) >= CONFCACHESIZE-1) return 0;
	ln[lines++]=bp;
	memmove(bp, p1, l+1);
	if(l<LINELEN_MIN) l=LINELEN_MIN; bp+=l+1;
    }
	/* overflow: returning 0 provokes infinite loop! */
    if(lines>=CONFLINES-1) {
	logprefix("CONF ERROR: line number overflow\n");
	lines=0; goto checklines;
    }
    if(bp>=confcache+CONFCACHESIZE) return 0;	/* cache overflow: need reinit. */
checklines:
    type[lines]=conftype_empty;
    ln[lines++]=empty=bp; *bp++=0;
    cache_end=bp;
    cl=getccache(lines*sizeof(struct ccline),sizeof(long));
    if(cl==NULL) return 0;
    for(l=inblock=0; l<lines && blocks<CONFBLOCKS; l++) {
	cl[l].type=type[l]; cl[l].name=ln[l]; cl[l].def=empty;
	switch(type[l]) {
	    case conftype_empty: {
		inblock=0;
		continue;
	    }
	    case conftype_path:
	    case conftype_path2: {
		cl[l].hash=0;
		for(p1=cl[l].name; *p1; p1++) {
		    if(*p1=='%' && p1[1]) {p1++; continue;}
		    if(*p1>='a' && *p1<='z') cl[l].hash |= 1<<(*p1-'a');
		}
		cl[l].len=strlen(cl[l].name);
		break;
	    }
	    case conftype_pre:
	    case conftype_def: {
		p1=strchr(ln[l],'='); if(p1==NULL) {
		    cl[l].name[0]=255; cl[l].def=empty;
		}
		else {
		    if(p1>ln[l] && p1[-1]=='!') {
			p1[-1]=0; negate=0x80;
		    }
		    else negate=0;
		    *p1++=0;
		    for(p2=p1; *p2; p2++) {
			if(*p2==',' || *p2==';' || *p2==':' || *p2=='@') *p2=' ';
		    }
		    strip_trailing_spaces(ln[l]); strip_trailing_spaces(p1);
		    i=search_list(cdef,cdefcnt,sizeof(cdef[0]),ln[l]);
		    if(i<0) i=255; cl[l].name[0]=i|negate;
		    p1=find_word_start(p1);
		    i=strlen(p1); memmove(cl[l].name+1,p1,i+1);
		    cl[l].def=cl[l].name+1;
		}
		break;
	    }
	    default: break;
	}
	if(!inblock) {
	    blk[blocks++].lines=&cl[l]; inblock=1;
	}
    }
	/* overflow: returning 0 provokes infinite loop! */
    if(blocks>=CONFBLOCKS) {
	logprefix("CONF ERROR: block number overflow\n");
	blocks=0;
    }
    blk[blocks++].lines=NULL;
    cf->blocks=getccache(blocks*sizeof(struct ccblock),sizeof(long));
    if(!cf->blocks) return 0;
    memmove(cf->blocks,blk,blocks*sizeof(struct ccblock));
    return 1;
}

struct cfp *readconf(int token, int type)
{
    char buf[SMK_PATH_MAX+1];
    char nbuf[32];
    const char *c, *comp[16];
    char cbuf[CONFLEN];
    struct stat st;
    struct cfp *cf;
    struct ccp *ccprep;
    char *old_cache_end;
    time_t mtime;
    int i, k, tryuser;
    int inited;	/* prevent infinite loop */

    tryuser=-1; inited=0;
    if(token&0xFFFF0000) {
	c="/user/";
	for(i=0; i<usercnt && users[i].uid!=request.uid; i++);
	if(i<usercnt && users[i].home && users[i].home[0]) tryuser=i;
    }
    else c="/sys/";
    regetcache:
    old_cache_end=cache_end;
    if(tryuser>=0) k=search_ccache(token,request.uid); else k=-1;
    if(k<0) k=search_ccache(token,-1);
    if(k<0) {
	force_getcache:
	if(cccnt>=CONFCACHELIM) goto reinit;
	ccprep=getccache(sizeof(struct ccp),sizeof(long));
	if(ccprep==NULL) {
	    reinit:
	    if(inited > 5) {	/* internal error */
		logprefix("BUG readconf(): cache init infinite loop.\n");
		return NULL;
	    }
	    inited++; confcache_init(); goto regetcache;
	}
	memset(ccprep,0,sizeof(struct ccp));
	new_confcache=1;
    }
    else {ccprep=cctab[k].cp;}
    switch(type) {
	case conf_mask: cf=&ccprep->mask; comp[3]="/mask"; break;
	case conf_open: cf=&ccprep->open; comp[3]="/open"; break;
	case conf_exec: cf=&ccprep->exec; comp[3]="/exec"; break;
	case conf_sock: cf=&ccprep->socket; comp[3]="/socket"; break;
	case conf_call: cf=&ccprep->call; comp[3]="/call"; break;
	/* case conf_trig: cf=&ccprep->trigger; comp[3]="/trigger"; break; */
	default: return NULL;
    }
    if(cf->laststat>now-STAT_DELAY) return cf;
    cf->laststat=now;
    int2str(token&0xFFFF,nbuf,sizeof(nbuf)); comp[2]=nbuf; comp[4]=NULL;
    if(tryuser>=0) {
	const char *save;
	comp[0]=users[tryuser].home; comp[1]="/.sysmask/";
	save=comp[3]; comp[3]=NULL;
	if(makestr(buf,comp,sizeof(buf))) return NULL;
	comp[3]=save;
	if(stat(buf,&st)==0 && S_ISDIR(st.st_mode)) {
		/* fix access rights */
	    if(st.st_mode & (S_IRWXO|S_IRWXG)) chmod(buf,0700);
	    if(k>=0 && cctab[k].user==-1) {	/* user def covers sys def */
		if(usercquota[tryuser].tquota >= USERTQUOTA ||
		   usercquota[tryuser].cquota >= USERCQUOTA)
		  goto sysdef;
		k=-1; goto force_getcache;
	    }
	    if(k<0) {	/* add tab */
		if(usercquota[tryuser].tquota >= USERTQUOTA ||
		   usercquota[tryuser].cquota >= USERCQUOTA)
		  goto sysdef;
		usercquota[tryuser].tquota++;
		k=cctab_insert(ccprep,token,request.uid);
	    }
	    goto thisfile;
	}
	else if(k>=0 && cctab[k].user>=0) {	/* user conf disappeared */
	    tryuser=-1; goto regetcache;
	}
    }
    sysdef:
    comp[0]=confdir; comp[1]=c; tryuser=-1;
    if(k<0) k=cctab_insert(ccprep,token,-1);
    thisfile:
    if(k<0) goto reinit;
    if(makestr(buf,comp,sizeof(buf))) return NULL;
    if(stat(buf,&st)) return cf;
	/* Fix the access rights if necessary */
    if(denyid && tryuser==-1 && st.st_uid!=denyid) chown(buf,denyid,denyid);
    if(tryuser>=0 && st.st_mode&(S_IRWXO|S_IRWXG)) chmod(buf,0600);
    if(st.st_size >= CONFLEN) return cf; /* oversized file refused */
    mtime=st.st_mtime;
	/* modified: erase the whole cache.
	 * The latency is there to prevent DoS attacks by users */
    if(!inited && mtime>=cache_time && mtime<now && 
       (tryuser<=0 || mtime<now-STAT_DELAY)) goto reinit;
	/* Cache overflow: restart it. */
    if(2*st.st_size + (cache_end-confcache) >= CONFCACHESIZE) goto reinit;
    if(!cf->blocks) {	/* missing in cache */
	cbuf[0]=0;
	readfile(buf,cbuf,sizeof(cbuf));
	if(!cutconf(cbuf, cf)) goto reinit;
	if(tryuser>=0) usercquota[tryuser].cquota+=cache_end-old_cache_end;
	new_confcache=1;
    }
    return cf;
}

	/* get definition from conf file, for request.fname.
	 * Returns 0 if no def found. */
int getdef(int token, int type)
{
    struct cfp *cf;
    struct ccline *cl;

    cf=readconf(token,type); if(!cf) return 0;
    cl=findf(cf); if(!cl) return 0;
    return readdef(cl);
}

	/* get and apply mask definition of the token */
int getmask(int token,struct sysmask_struct *sm)
{
    struct cfp *cf;
    
    cf=readconf(token,conf_mask);
    if(!cf || !cf->blocks || !cf->blocks->lines) return 0;
    readdef(cf->blocks->lines);
    def2mask(sm);
    return 1;
}

