/*    Copyright (C) 2005 XIAO, Gang of Universite de Nice - Sophia Antipolis
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

	/* socket translation for smkd */

#include <stdint.h>

#define UNIX_PATH_MAX 108

struct sockaddr_un
{
    sa_family_t family;
    char sun_path[UNIX_PATH_MAX];	/* Path name.  */
};

typedef uint32_t in_addr_t;
typedef uint16_t in_port_t;
struct in_addr
  {
    in_addr_t s_addr;
  };

struct sockaddr_in
  {
    sa_family_t family;
    in_port_t sin_port;			/* Port number.  */
    struct in_addr sin_addr;		/* Internet address.  */
  };

void ipv4_str(struct in_addr *ipv4, char *buf, int buflen)
{
    char *p;
    unsigned char *v;
    int i;
    v=(unsigned char *) &ipv4->s_addr;
    buf[0]=0; 
    for(i=0, p=buf; i<4 && p-buf<buflen-5; i++, p+=strlen(p)) {
	if(i) *p++='.';
	int2str(v[i],p,buflen-(p-buf));
    }
}

void port_str(in_port_t port, char *buf, int buflen)
{
    int i;
    
    i=((port>>8)&255) | ((port<<8)&0xFF00);
    int2str(i,buf,buflen);
}

	/* returns 1 if ok */
int sock2str(void)
{
    int reqtype, family, socktype;
    int *ip;
    void *p;
    const char *sm[32];
    char namebuf[SMK_PATH_MAX+1];
    char portbuf[32];
    struct sockaddr_un *addr_un;
    struct sockaddr_in *addr_in;
    
    request.fname[request.reqlen]=0;
    ip=(int *) request.fname;
    reqtype=ip[0]; family=ip[1]; socktype=ip[2];
    p=request.fname+3*sizeof(int);
    switch(reqtype) {
	case SYS_BIND: sm[0]="/bind/"; break;
	case SYS_LISTEN: sm[0]="/listen/"; break;
	case SYS_ACCEPT: sm[0]="/accept/"; break;
	case SYS_CONNECT: sm[0]="/connect/"; break;
	case SYS_SETSOCKOPT: sm[0]="/setsockopt/"; break;
	case SYS_SEND: sm[0]="/send/"; break;	/* non-empty address only */
	/*case SYS_RECV: sm[0]="/recv/"; break;*/		/* non-used */
	/*case SYS_GETSOCKOPT: sm[0]="/getsockopt/"; break;*/	/* non-used */
	default: ans.ind=EINVAL; return 0;
    }
    switch(family) {
	case PF_UNIX: {
	    sm[1]="local";
	    addr_un=p;
	    if(addr_un->family!=AF_UNIX) {
		sm[1]="local"; sm[2]="/ERROR";
	    }
	    else {
		sm[2]=addr_un->sun_path;
		addr_un->sun_path[UNIX_PATH_MAX-1]=0;
		if(sm[2][0]!='/') sm[2]="";
	    }
	    sm[3]=NULL;
	    break;
	}
	case PF_INET: {
	    switch(socktype) {
		case SOCK_STREAM: sm[1]="tcp/"; break;
		case SOCK_DGRAM: sm[1]="udp/"; break;
		default: sm[1]="ipv4/"; break;
	    }
	    if(request.reqlen<=3*sizeof(int)) {
		sm[2]=NULL; break;
	    }
	    addr_in=p;
	    ipv4_str(&addr_in->sin_addr,namebuf,sizeof(namebuf));
	    port_str(addr_in->sin_port,portbuf,sizeof(portbuf));
	    sm[2]=namebuf; sm[3]="/"; sm[4]=portbuf; sm[5]=NULL;
	    break;
	}
	default: ans.ind=EACCES; return 0;
    }
    if(makestr(request.fname,sm,SMK_PATH_MAX)) {
	ans.ind=ENAMETOOLONG; return 0;
    }
    process_pathname(0);
    return 1;
}

